/*
This firmware was built for Geoff's flipbot remote control, although
the functionality has been kept platform-agnostic for compatibility
with future remote-controlled devices. The transmitted value from
the remote is a single byte containing the x and y values of the
joystick, each ranging from 0 to 15 (for a total of 4 bits per
axis). The x axis is the MSB 4 bits, and the y axis is the LSB 4 bits.

Because there is no integer median in [0, 15], we consider both
7 and 8 to be neutral positions.

See the 'flipbot' directory for more information and for the robot
firmware.

To access debug data, plug into a USB port and open a virtual COM port
at 9600 baud.
*/

#include <wixel.h>
#include <usb.h>
#include <usb_com.h>
#include <radio_queue.h>

#include <stdio.h>

#define TX_PERIOD 1

uint16 lastTxTime = 0;

void updateLEDs()
{
	LED_YELLOW((getMs() & 0x3FF) <= 20);	// heartbeat
	LED_RED(0);
	usbShowStatusWithGreenLed();
}

// for printf
void putchar(char c)
{
	usbComTxSendByte(c);
}

// determine if a command represents a neutral stick
uint8 isNeutralCommand(uint8 command)
{
	uint8 x = (command >> 4);
	uint8 y = (command & 0xff);
	return (x == 7 || x == 8) && (y == 7 || y == 8);
}

// returns a byte representing the state of the joystick
uint8 readJoystickCommand()
{
	const uint16 MAX_ANALOG_READ = 2047;
	const uint16 ANALOG_TO_4BITS = 128;
	int16 analog0, analog1;
	uint16 x, y;
	
	// read the analog voltage from the joystick and then
	// scale it to fit in four bits per axis (the raw analog
	// voltage values range from 0 to roughly 3356 millivolts)
	adcSetMillivoltCalibration(adcReadVddMillivolts());
	analog0 = adcRead(0) - 82;		// we apply these adjustments because testing shows that
	analog1 = adcRead(1) - 92;		// the voltage range isn't exactly 0 to 3356
	
	// this shouldn't be necessary, but we want to be sure that
	// the joystick readings are positive
	analog0 = (analog0 < 0 ? 0 : analog0);
	analog1 = (analog1 < 0 ? 0 : analog1);
	
	// scale the joystick analog readings to a 4-bit range;
	// the x-axis is reversed because of the orientation of
	// the joystick
	x = ((MAX_ANALOG_READ - analog1) / ANALOG_TO_4BITS);
	y = (analog0 / ANALOG_TO_4BITS);

	// USB debugging output
	if (usbPowerPresent() && usbComTxAvailable() >= 45)
	{
		printf("a1: %04u, a0: %04u, x: %02u, y: %02u\r\n", analog1, analog0, x, y);
	}
	
	// the resulting command is stored in a single byte
	return (x << 4) | (y & 0x0f);
}

// transmits a joystick state byte via radio
void rcService()
{
	uint8 XDATA *txPacket;
	uint8 command;
	
	// this delay is currently negligible, but we'll keep it in case we need it in the future
	if((uint16)(getMs() - lastTxTime) >= TX_PERIOD)
	{
		// read the command from the joystick and prepare the TX packet
		command = readJoystickCommand();
		
		// there's no need to send bytes that represent a neutral stick
		if(!isNeutralCommand(command))
		{
			txPacket = radioQueueTxCurrentPacket();
		
			// ensure that we have a free packet to transmit
			if(txPacket)
			{
				lastTxTime = getMs();
				txPacket[0] = 1;
				txPacket[1] = command;
				radioQueueTxSendPacket();
			}
		
		}
	}
}

void main()
{
	systemInit();
	usbInit();
	radioQueueInit();
	
	while(1)
	{
		updateLEDs();
        boardService();
        usbComService();			// required for programming
        rcService();				// controls motors using remote control commands
    }
}
